/*
 * loops.c
 *
 * CS 470 Activity
 * Original serial version.
 *
 * Compile with --std=c99
 */

#include <math.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>

#ifdef _OPENMP
#   include <omp.h>
#   define START_TIMER(X)   float X = omp_get_wtime();
#   define END_TIMER(X)     X = (omp_get_wtime()-(X));
#   define CONVERT_TIMER(X) (X)
#else
#   define START_TIMER(X)   clock_t X = clock();
#   define END_TIMER(X)     X = (clock()-(X));
#   define CONVERT_TIMER(X) ((float)(X)/(float)CLOCKS_PER_SEC)
#endif

/*
 * Returns true if x is a prime number and false otherwise.
 */
bool is_prime(int x)
{
    if (x % 2 == 0) {       // eliminate even numbers
        return false;
    }
    for (int d = 3; d < (int)ceil(sqrt((float)x)); d += 2) {
        if (x % d == 0) {   // naive test
            return false;
        }
    }
    return true;
}

/*
 * Returns true if the given array is sorted and false otherwise.
 */
bool is_sorted(double *data, int len)
{
    for (int i = 0; i < len-1; i++) {
        if (data[i] > data[i+1]) {
            return false;
        }
    }
    return true;
}


int main(int argc, char* argv[])
{
    srand((unsigned)time(NULL));

    /*
     * sum of primes
     */
    START_TIMER(prime_sum)
    double psum = 0.0;
    for (int i = 1; i < 5000000; i++) {
        if (is_prime(i)) {
            psum += (double)i;
        }
    }
    END_TIMER(prime_sum)
    printf("prime_sum: %.4f %.6e\n", CONVERT_TIMER(prime_sum), psum);

    const int DSIZE = 250000000;
    //const int DSIZE = 50000;        // use for sort testing

    double *data = (double*)malloc(sizeof(double) * DSIZE);

    /*
     * data initialization (square roots)
     */
    START_TIMER(data_init)
    for (int i = 0; i < DSIZE; i++) {
        data[i] = sqrt((double)i);
    }
    END_TIMER(data_init)
    printf("data_init: %.4f\n", CONVERT_TIMER(data_init));

    /*
     * data swapping
     */
    START_TIMER(data_swap)
    for (int i = 0; i < DSIZE/2; i++) {
        double tmp = data[i];
        data[i] = data[DSIZE-i];
        data[DSIZE-i] = tmp;
    }
    END_TIMER(data_swap)
    printf("data_swap: %.4f\n", CONVERT_TIMER(data_swap));

#if 0
    /*
     * data shuffling
     */
    START_TIMER(data_shuf)
    for (int i = DSIZE-1; i > 0; i--) {
        int j = rand() % (i+1);
        double tmp = data[i];
        data[i] = data[j];
        data[j] = tmp;
    }
    END_TIMER(data_shuf)
    printf("data_shuf: %.4f\n", CONVERT_TIMER(data_shuf));
#endif

    /*
     * data sorting
     */
#if 0
    START_TIMER(data_sort)
    for (int i = 0; i < DSIZE-1; i++) {
        int min = i;
        double minval = data[i];
        for (int j = i; j < DSIZE; j++) {
            if (data[j] < minval) {
                min = j;
                minval = data[j];
            }
        }
        double tmp = data[i];
        data[i] = data[min];
        data[min] = tmp;
    }
    END_TIMER(data_sort)
    printf("data_sort: %.4f %s\n", CONVERT_TIMER(data_sort),
            (is_sorted(data, DSIZE) ? "pass" : "FAIL"));
#endif

    /*
     * data summation
     */
    START_TIMER(data_sum)
    double dsum = 0.0;
    for (int i = 0; i < DSIZE; i++) {
        dsum += data[i];
    }
    END_TIMER(data_sum)
    printf("data_sum:  %.4f %.6e\n", CONVERT_TIMER(data_sum), dsum);

    printf("\n");
    free(data);
    return EXIT_SUCCESS;
}

